(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`AtomicSystem`", {"AtomicDensityMatrix`"}];

Begin["`Private`"]; (* Begin Private Context *)

(* ::Subsection:: *)
(*Atomic system*)

(*AtomicState[ label ] specifies an atomic state labeled by label*)
(*AtomicState[ label, {\!\(TraditionalForm\`\**)
(*SubscriptBox[*)
(*StyleBox["param",*)
(*FontSlant->"Italic"], "1"]\)->Subscript[value, 1] , ...} ] specifies an atomic state with one or more parameters set*)
(**)
(*Level parameters:*)
(*	S		electronic spin quantum number*)
(*	L		electronic orbital angular momentum quantum number*)
(*	J		total electronic angular momentum quantum number*)
(*	NuclearSpin	nuclear spin quantum number*)
(*	F		total angular momentum quantum number*)
(*	M		z-projection of total angular momentum*)
(*	Parity		parity*)
(*	Energy		unperturbed energy*)
(*	NaturalWidth	natural line width*)
(*	BranchingRatio[label]	branching ratio to state designated by label*)
(*	LevelPosition	"temporary" parameter that specifies whether the state is upper or lower for the purposes of a spontaneous decay calculation*)

AtomicState[label_, {rules___Rule}] := AtomicState[label, rules];

(*StateLabel[state] gets the label from a state specification*)

StateLabel[state_AtomicState] := First[state];
StateLabel[sys_?AtomicSystemQ] := StateLabel /@ sys;

(*StateParameters[state] gets the parameters from a state specification*)

StateParameters[state_AtomicState] ^:= Rest[List @@ state];

(*MapOnParameters[f_, state_AtomicState] := AtomicState[StateLabel[state], f[StateParameters[state]]];*)

(*StateParameterPattern[] matches valid atomic state parameters*)

StateParameterPattern[] :=
    S | L | J | NuclearSpin | F | M | Parity | Energy | NaturalWidth |
        GFactor | BranchingRatio[_] | LevelPosition | HyperfineA |
        HyperfineB | Polarizability[_] | TransitionPolarizability[_];

(*StateParameterQ[expr] tests if an expression is a valid atomic state parameter*)

StateParameterQ[expr_] := MatchQ[expr, StateParameterPattern[]];

(*AtomicSystemQ[expr] tests if an expression is a list of AtomicState specifications.*)

AtomicSystemPattern[] = {_AtomicState..};

AtomicSystemQ[expr_] := MatchQ[expr, AtomicSystemPattern[]];

(*DefinedQ[state,param] checks if an atomic state parameter is defined in a state specification.*)

DefinedQ[state_AtomicState, x : StateParameterPattern[]] := MemberQ[Rest[state], x -> _];

DefinedQ[sys : _AtomicState | {_AtomicState ..}, x : {StateParameterPattern[] ..} | StateParameterPattern[] ..] :=
    And @@ Flatten@Outer[DefinedQ, Flatten@{sys}, Flatten@{x}];

(*DeleteParameter[state, param] deletes a rule from a state specification. More than one parameter can be deleted by giving a list of parameters or additional arguments.*)

DeleteParameter[sys : {_AtomicState ..} | _AtomicState, x : {StateParameterPattern[] ..} | StateParameterPattern[]..] :=
    DeleteParameter[sys, x] =
        DeleteCases[sys, Alternatives @@ Thread[Flatten[{x}] -> _], {ArrayDepth[{sys}]}];

(*AddParameter[state, param -> value] adds a rule to an atomic state specification. If the parameter is already present, its value is changed if necessary.*)

AddParameter[state_AtomicState, rules : {(StateParameterPattern[] -> _) ..} | (StateParameterPattern[] -> _)..] :=
    Join[DeleteParameter[state, Flatten[{rules}][[All, 1]]], AtomicState @@ Flatten[{rules}]];
AddParameter[sys_?AtomicSystemQ, rules : {(StateParameterPattern[] -> _) ..} | (StateParameterPattern[] -> _)..] :=
    Thread[Unevaluated@AddParameter[sys, rules], List, 1];

(*param[state] returns the parameter for a state if defined, otherwise returns a placeholder marked by the state label.*)
(*It caches results for improved performance.*)

(*(#[state_AtomicState]:=#/.StateParameters[state]/.#->#[StateLabel[state]])&/@{S,L,NuclearSpin,M,Parity,Energy,NaturalWidth,GFactor,LevelPosition,HyperfineA,HyperfineB};*)

With[{f = #},
    f[state_AtomicState] := f[state] = f /. StateParameters[state] /. f -> f[StateLabel[state]];
    f[sys_?AtomicSystemQ] := f /@ sys;
]& /@ {S, L, NuclearSpin, M, Parity, Energy, NaturalWidth, GFactor, LevelPosition, HyperfineA, HyperfineB};

(*J[state] returns the J for a state if defined, otherwise returns 0*)

J[state_AtomicState] := J[state] = J /. StateParameters[state] /. J -> 0;
J[sys_?AtomicSystemQ] := J /@ sys;

(*F[state] returns F for a state if defined, otherwise returns J*)

F[state_AtomicState] := F[state] = F /. StateParameters[state] /. F -> J[state];
F[sys_?AtomicSystemQ] := F /@ sys;

BranchingRatio[state1_AtomicState][state2_AtomicState] := If[
    (*if reverse branching ratio is nonzero, or states have the same parity, then this b.r. is zero*)
    (DefinedQ[state1, BranchingRatio[StateLabel@state2]] && BranchingRatio[StateLabel@state2][state1] =!= 0)
        || Parity[state1] === Parity[state2],
    0,
    (*else calculate b.r. using state label*)
    BranchingRatio[StateLabel[state1]][state2]
];

BranchingRatio[label_][state_AtomicState] := Module[{value},
    If[
        StateLabel[state] === label || NaturalWidth[state] === 0,
        0,
        If[
            (value = BranchingRatio[label] /. StateParameters[state]) =!= BranchingRatio[label],
            value,
            BranchingRatio[StateLabel[state], label]
        ]
    ]
];

Polarizability[k__][state_AtomicState] := Module[{value},
    If[(value = Polarizability[k] /. StateParameters[state]) =!= Polarizability[k],
        value,
        Polarizability[k][StateLabel[state]]
    ]
];

TransitionPolarizability[k_][state1_AtomicState, state2_AtomicState] := Module[{value},
    Which[
        (value = TransitionPolarizability[k][StateLabel[state1]] /. StateParameters[state2]) =!= TransitionPolarizability[k][StateLabel[state1]], value,
        (value = TransitionPolarizability[k][StateLabel[state2]] /. StateParameters[state1]) =!= TransitionPolarizability[k][StateLabel[state2]], value,
        True, TransitionPolarizability[k][Sequence @@ Sort[StateLabel /@ {state1, state2}]]
    ]
];

GroundStateQ[state_AtomicState] := NaturalWidth[state] === 0;
SetAttributes[GroundStateQ, Listable];

(*GroundStates[system] returns all ground states.*)

GroundStates[sys_?AtomicSystemQ] := Select[sys, GroundStateQ];

(*ExcitedStates[sys] returns all not ground states.*)

ExcitedStates[sys_?AtomicSystemQ] := Select[sys, !GroundStateQ[#]&];

SublevelMultiplicity[s_AtomicState] :=
    If[DefinedQ[s, F],
        (2F[s] + 1),
        If[DefinedQ[s, J], (2 J[s] + 1), 1]If[DefinedQ[s, NuclearSpin], (2 NuclearSpin[s] + 1), 1]
    ];

FineStructureStatesQ[s1_AtomicState, s2_AtomicState] :=
    ListQ@StateLabel@s1 && ListQ@StateLabel@s2 && First@StateLabel@s1 === First@StateLabel@s2

SublevelMultiplicity[sys_?AtomicSystemQ] :=
    Total[SublevelMultiplicity /@ (DeleteParameter[SelectState[sys, #], F]& /@ Union[StateLabel[sys]])];

(*SelectStates[states, rule1, rule2, ...] returns all states that satisfy the supplied rules*)

(*propMatch[a_->b_,reg_]:=MatchQ[a[reg],b]*)
(*propMatch[expr_,reg_]:=expr/.Rule[a_,b_]:>MatchQ[a[reg],b]
critFunc[crit__][reg_]:=And@@(propMatch[#,reg]&/@{crit})*)
(*SelectStates[sys:{_AtomicState..},crit__]:=Select[sys,critFunc[crit]]
SelectStates[sys:{_AtomicState..},crit__,n_?IntegerQ]:=Take[Select[sys,critFunc[crit]],n]*)
(*DeleteStates[sys:{_AtomicState..},crit__]:=DeleteCases[sys,s_?(critFunc[crit])]*)
(*StatePosition[sys:{_AtomicState..},crit__]:=Position[sys,s_?(critFunc[crit])]*)
(*StatePosition[sys:{_AtomicState..}, crit__] :=
 Position[Transpose@Outer[#1[#2]&, {crit}[[All, 1]], sys], {crit}[[All, 2]]]*)

SelectStates[sys_?AtomicSystemQ, crit_, n_ : Infinity] := Extract[sys, StatePosition[sys, crit, n]]

SelectState[sys_?AtomicSystemQ, crit_] := First@SelectStates[sys, crit, 1];

DeleteStates[sys_?AtomicSystemQ, crit_] := Delete[sys, StatePosition[sys, crit]];

StateLabelQ[state_AtomicState, expr_] := StateLabel[state] === expr;

StateLabelQ[sys_?AtomicSystemQ, expr_] := MemberQ[StateLabel[sys], expr];

StateEvaluate[state_AtomicState, expr_] := expr /. param : (StateParameterPattern[] | StateLabel) :> param[state];

StateEvaluate[sys_?AtomicSystemQ, expr_] := StateEvaluate[#, expr]& /@ sys;

StatePosition[sys_?AtomicSystemQ, crit_, n_ : Infinity] /; StateLabelQ[sys, crit] :=
    StatePosition[sys, StateLabel == crit, n];

StatePosition[sys_?AtomicSystemQ, crit_, n_ : Infinity]  /; !StateLabelQ[sys, crit] :=
    Position[sys, state_ /; StateEvaluate[state, crit], {1}, n, Heads -> False];

StateMatchQ[state_AtomicState, crit_] /; StateLabelQ[state, crit] := True;

StateMatchQ[state_AtomicState, crit_] := TrueQ@StateEvaluate[state, crit];

StateMatchQ[sys_?AtomicSystemQ, crit_] := StateMatchQ[#, crit]& /@ sys;

MatrixElementPosition[sys_?AtomicSystemQ, crit1_ -> crit2_, n_ : Infinity] :=
    Module[{pos1, pos2},
        {pos1, pos2} = Flatten@StatePosition[sys, #, n] & /@ {crit1, crit2};
        Union[#, Reverse /@ #]&@Flatten[Outer[List, pos1, pos2], 1]
    ];
MatrixElementPosition[sys_?AtomicSystemQ, crit : {__Rule}] :=
    Union @@ (MatrixElementPosition[sys, #] &) /@ crit;
MatrixElementPosition[sys_?AtomicSystemQ, All | True] :=
    Tuples[Range@Length@sys, 2];
MatrixElementPosition[sys_?AtomicSystemQ, None | False] := {};

MatrixElements[sys_?AtomicSystemQ, a_?OperatorQ, crit_Rule] := Extract[a, MatrixElementPosition[sys, crit]];

MatrixElement[sys_?AtomicSystemQ, a_, {s1_AtomicState, s2_AtomicState}] :=
    Extract[a, Flatten[Position[sys, #, {1}, 1, Heads -> False]& /@ {s1, s2}]];

MatrixElement[sys_?AtomicSystemQ, a_, crit_Rule] := First@Extract[a, MatrixElementPosition[sys, crit, 1]];

AtomicSystemType::difftypes = "All atomic states in the system are not of the same type, `1`, as the first state.";
AtomicSystemType::unrec = "The type of the atomic state `1` is not recognized.";

ValidQuantumNumberQ[s_AtomicState, j_] := DefinedQ[s, j] && HalfIntegerQ@j@s;

AtomicSystemType[s_AtomicState] :=
    Module[{noMessage = True, result},
        result = Check[
            Module[{fDefined, jDefined, mDefined},
                {fDefined, jDefined, mDefined} = ValidQuantumNumberQ[s, #]& /@ {F, J, M};
                Which[
                    fDefined && jDefined && mDefined,
                    "HyperfineZeeman",
                    fDefined && jDefined && !mDefined,
                    "Hyperfine",
                    !fDefined && jDefined && mDefined,
                    "Zeeman",
                    jDefined && !mDefined,
                    "Fine",
                    !jDefined && !mDefined,
                    "Toy",
                    True,
                    Message[AtomicSystemType::unrec, s]
                ]
            ],
            noMessage = False
        ];
        (AtomicSystemType[s] = result) /; noMessage
    ];

AtomicSystemType[sys_?AtomicSystemQ] :=
    Module[{noMessage = True, result},
        result = Check[
            With[{types = AtomicSystemType /@ sys},
                If[MatchQ[Hold@types, Hold@{_String..}],
                    If[MatchQ[types, {First@types..}],
                        First@types,
                        Message[AtomicSystemType::difftypes, First@types]
                    ]
                ]
            ],
            noMessage = False
        ];
        (AtomicSystemType[sys] = result) /; noMessage
    ];

SetAttributes[LevelLabel, Listable];
LevelLabel[s_AtomicState] := Switch[AtomicSystemType@s,
    "Toy" | "Zeeman" | "Fine",
    StateLabel@s,
    "HyperfineZeeman" | "Hyperfine",
    {StateLabel@s, F@s}
];

SetAttributes[SublevelLabel, Listable];
SublevelLabel[s_AtomicState] := Switch[AtomicSystemType@s,
    "Toy" | "Fine",
    StateLabel@s,
    "Zeeman",
    {StateLabel@s, M@s},
    "Hyperfine",
    {StateLabel@s, F@s},
    "HyperfineZeeman",
    {StateLabel@s, F@s, M@s}
];

HyperfineSublevels[s_AtomicState] := If[
    AtomicSystemType@s == "Fine" && DefinedQ[s, NuclearSpin] && HalfIntegerQ@NuclearSpin@s,
    Reverse@Table[AddParameter[s, F -> f], {f, Abs[NuclearSpin@s - J@s], NuclearSpin@s + J@s}],
    s
];

HyperfineSublevels[sys_?AtomicSystemQ] :=
    Flatten[HyperfineSublevels /@ sys];

ZeemanSublevels[s_AtomicState] :=
    Switch[AtomicSystemType@s,
        "Fine",
        Reverse@Table[AddParameter[s, M -> m], {m, -J@s, J@s}],
        "Hyperfine",
        Reverse@Table[AddParameter[s, M -> m], {m, -F@s, F@s}],
        _,
        s
    ];

ZeemanSublevels[sys_?AtomicSystemQ] :=
    Flatten[ZeemanSublevels /@ sys];

Sublevels[s_AtomicState] := ZeemanSublevels@HyperfineSublevels@s;

Sublevels[sys_?AtomicSystemQ] := Flatten[Sublevels /@ sys];

RemoveZeemanSublevels[sys_?AtomicSystemQ] := First /@ Split@DeleteParameter[sys, M];

CountZeemanSublevels[sys_?AtomicSystemQ] := Length /@ Split@DeleteParameter[sys, M];

(*ZeemanSystem[label, J] constructs a spin J atomic system.*)

ZeemanSystem[label_, j_] := ZeemanSublevels[AtomicState[label, J -> j]];

ZeemanSystem[j_] := ZeemanSystem[ToString@Unique[], j];

End[]; (* End Private Context *)

EndPackage[];
