(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

(*The MatrixRSolve package aids the generation and numerical solution of matrix recursion relations.*)


BeginPackage["AtomicDensityMatrix`MatrixRSolve`", {"AtomicDensityMatrix`"}];

(* ::Subsection:: *)

Begin["`Private`"];

(* ::Subsection::Closed:: *)
(*MatrixRSolve*)

(* ::Text:: *)
(*Numerically solve matrix recursion relations of the form*)
(*	\!\( *)
(*\*SubsuperscriptBox[\(C\), \(n\), \(-\)]\ *)
(*\*SubscriptBox[\(a\), \(n - 1\)]\)+Subscript[C, n] Subscript[a, n]+\!\( *)
(*SubsuperscriptBox[\(C\), \(n\), \(+\)] *)
(*\*SubscriptBox[\(a\), \(n + 1\)]\)+c Subscript[\[Delta], n,0]=0,*)
(*where Subscript[a, -n]=\!*)
(*\*SubsuperscriptBox[\(a\), \(n\), \( *\)].*)

(* ::Text:: *)
(*We want matrices Subscript[S, n] such that *)
(*	Subscript[a, n]=Subscript[S, n] Subscript[a, n-1].*)
(*Then for n>0,*)
(*	0=\!\( *)
(*\*SubsuperscriptBox[\(C\), \(n\), \(-\)]\ *)
(*\*SubscriptBox[\(a\), \(n - 1\)]\)+Subscript[C, n] Subscript[S, n] Subscript[a, n-1]+\!\( *)
(*SubsuperscriptBox[\(C\), \(n\), \(+\)] *)
(*SubscriptBox[\(S\), \(n + 1\)] *)
(*SubscriptBox[\(S\), \(n\)] *)
(*\*SubscriptBox[\(a\), \(n - 1\)]\)*)
(*   =(\!\( *)
(*\*SubsuperscriptBox[\(C\), \(n\), \(-\)] + \( *)
(*SubscriptBox[\(C\), \(n\)] *)
(*\*SubscriptBox[\(S\), \(n\)]\) + \( *)
(*SubsuperscriptBox[\(C\), \(n\), \(+\)] *)
(*SubscriptBox[\(S\), \(n + 1\)] *)
(*\*SubscriptBox[\(S\), \(n\)]\)\)) Subscript[a, n-1] ,*)
(*which is true if*)
(*	Subscript[S, n]=-(Subscript[C, n]+\!\( *)
(*SubsuperscriptBox[\(C\), \(n\), \(+\)] *)
(*\*SubscriptBox[\(S\), \(n + 1\)]\))^-1 \!*)
(*\*SubsuperscriptBox[\(C\), \(n\), \(-\)].*)

SIterator[{Splus_, nplus_}, {CMatrix_, CPlus_, CMinus_}] :=
    Module[{n = nplus - 1}, {-Inverse[CMatrix[n] + CPlus[n].Splus].CMinus[n], n}];

(* ::Text:: *)
(*We set Subscript[S, m]=0 for some m, and calculate Subscript[S, m-1],... ,Subscript[S, 1] *)

SList[{CMatrix_, CPlus_, CMinus_}, m_] :=
    Reverse[NestList[SIterator[#, {CMatrix, CPlus, CMinus}]&, {0CMatrix[m], m}, m - 1][[All, 1]]];

(* ::Text:: *)
(*The zeroth coefficent Subscript[a, 0]is given by*)
(*	\!\( *)
(*\*SubsuperscriptBox[\(C\), \(0\), \(-\)]\ *)
(*\*SubscriptBox[\(a\), \(-1\)]\)+Subscript[C, 0] Subscript[a, 0]+\!\( *)
(*SubsuperscriptBox[\(C\), \(0\), \(+\)] *)
(*\*SubscriptBox[\(a\), \(1\)]\)+c=0*)
(*	\!\( *)
(*\*SubsuperscriptBox[\(C\), \(0\), \(-\)]\ *)
(*\*SubsuperscriptBox[\(a\), \(1\), \( *\)]\)+Subscript[C, 0] Subscript[a, 0]+\!\( *)
(*SubsuperscriptBox[\(C\), \(0\), \(+\)] *)
(*\*SubscriptBox[\(a\), \(1\)]\)+c=0*)
(*	(\!\( *)
(*\*SubsuperscriptBox[\(C\), \(0\), \(-\)]\ *)
(*\*SubsuperscriptBox[\(S\), \(1\), \( *\)]\)+Subscript[C, 0]+\!\( *)
(*SubsuperscriptBox[\(C\), \(0\), \(+\)] *)
(*\*SubscriptBox[\(S\), \(1\)]\)) Subscript[a, 0]=-c*)
(*or*)
(*	Subscript[a, 0]=-(\!\( *)
(*\*SubsuperscriptBox[\(C\), \(0\), \(-\)]\ *)
(*\*SubsuperscriptBox[\(S\), \(1\), \( *\)]\)+Subscript[C, 0]+\!\( *)
(*\*SubsuperscriptBox[\(C\), \(0\), \(+\)]\ *)
(*\*SubscriptBox[\(S\), \(1\)]\))^-1c.*)

aZero[S1_, {CMatrix_, CPlus_, CMinus_}, cVector_] :=
    -Inverse[CMinus[0].Conjugate[S1] + CMatrix[0] + CPlus[0].S1].cVector;

(* ::Text:: *)
(*Calculate Subscript[a, n] using Subscript[a, n]=Subscript[S, n] Subscript[a, n-1].*)

aIterator[{aMinus_, nMinus_}, SList_] := {SList[[nMinus + 1]].aMinus, nMinus + 1};

aList[matfuncs_, cVector_, SList_, n_] :=
    NestList[aIterator[#, SList]&, {aZero[First@SList, matfuncs, cVector], 0}, n][[All, 1]];

(* ::Text:: *)
(*Put it all together.*)

MatrixRSolve[{cmat_, cplus_, cminus_, cVector_}, {nvar_, n1_, n2_}] :=
    Module[{matfuncs},
        matfuncs = If[
            ListQ@cmat,
            Function /@ N[{cmat, cplus, cminus}] /. nvar -> Slot[1],
            N[{cmat, cplus, cminus}]
        ];
        aList[matfuncs, N[cVector], SList[matfuncs, n1], n2]
    ];

RecursionSolutionToReplacements[sol_, vars_, t_] :=
    Flatten[MapIndexed[MapThread[Rule, {vars /. t -> #2[[1]] - 1, #1}]&, sol]];

(* ::Subsection:: *)
(*MatrixRSolve generalized*)

(* ::Input:: *)
(*(*Numerically solve matrix recursion relations of the form *)
(*	\!*)
(*\*SubsuperscriptBox[\(C\), \(n\), \(-\)] Subscript[a, n-1]+Subscript[C, n] Subscript[a, n]+SubsuperscriptBox[C,n,+] Subscript[a, n+1]+c Subscript[\[Delta], n,0]=0,where  Subscript[a, -n]=\!*)
(*\*SubsuperscriptBox[\(a\), \(n\), \( *\)].*)*)

(* ::Input:: *)
(*(*We want matrices Subscript[S, n] such that*)
(*	Subscript[a, n]=Subscript[S, n] Subscript[a, n-1].*)
(*Then for  n>0,*)
(*	0=\!\( *)
(*\*SubsuperscriptBox[\(C\), \(n\), \(-\)]\ *)
(*\*SubscriptBox[\(a\), \(n - 1\)]\)+Subscript[C, n] Subscript[S, n] Subscript[a, n-1]+SubsuperscriptBox[C,n,+]SubscriptBox[S,n+1]SubscriptBox[S,n] Subscript[a, n-1]*)
(*	=(\!\( *)
(*\*SubsuperscriptBox[\(C\), \(n\), \(-\)] + \(SubscriptBox[C, n] *)
(*\*SubscriptBox[\(S\), \(n\)]\) + \(SubsuperscriptBox[C, n, +] SubscriptBox[S, n + 1] *)
(*\*SubscriptBox[\(S\), \(n\)]\)\)) Subscript[a, n-1],*)
(*which is true if *)
(*	Subscript[S, n]=-(Subscript[C, n]+SubsuperscriptBox[C,n,+] Subscript[S, n+1])^-1 \!*)
(*\*SubsuperscriptBox[\(C\), \(n\), \(-\)]. *)*)

SIterator[{Splus_, nplus_}, {CMinus_, CZero_, CPlus_}] :=
    Module[{n = nplus - 1}, {SparseArray[-LinearSolve[CZero[n] + CPlus[n].Splus, CMinus[n]]], n}];

(* ::Text:: *)
(*We set Subscript[S, m]=0 for some m, and calculate Subscript[S, m-1],... ,Subscript[S, 1] *)

SList[{CMinus_, CZero_, CPlus_}, m_] :=
    Reverse[NestList[SIterator[#, {CMinus, CZero, CPlus}]&, {SparseArray[{_, _} -> 0, Dimensions[CZero[m]]], m}, m - 1][[All, 1]]];

(* ::Text:: *)
(*The zeroth coefficent Subscript[a, 0]is given by*)
(*	Subscript[a, 0]=-(\!\( *)
(*\*SubsuperscriptBox[\(C\), \(0\), \(-\)]\ *)
(*\*OverscriptBox[*)
(*SubscriptBox[\(S\), \(1\)], \(~\)]\)+Subscript[C, 0]+\!\( *)
(*\*SubsuperscriptBox[\(C\), \(0\), \(+\)]\ *)
(*\*SubscriptBox[\(S\), \(1\)]\))^-1c.*)

TildeOp[mat_, nvars_] := SparseArray[Conjugate@ArrayFlatten@Reverse[Partition[mat, {nvars, nvars}], {1, 2}]];

aZero[S1_, {CMinus_, CZero_, CPlus_}, cVector_, nvars_] :=
    -LinearSolve[CMinus[0].TildeOp[S1, nvars] + CZero[0] + CPlus[0].S1, cVector];

(* ::Text:: *)
(*Calculate Subscript[a, n] using Subscript[a, n]=Subscript[S, n] Subscript[a, n-1].*)

aIterator[{aMinus_, nMinus_}, SList_] := {SList[[nMinus + 1]].aMinus, nMinus + 1};

aList[matfuncs_, cVector_, SList_, L_, nvars_, n2_] :=
    Take[
        Drop[
            Partition[
                Flatten@NestList[
                    aIterator[#, SList]&,
                    {aZero[First@SList, matfuncs, cVector, nvars], 0},
                    Ceiling[n2 / L]
                ][[All, 1]],
                nvars
            ],
            (L - 1) / 2
        ],
        n2 + 1
    ];

(* ::Text:: *)
(*Put it all together.*)

MakeMatrixFunctions[{CMinus_, CZero_, CPlus_}, nvar_] :=
    Module[{boo},
        If[MatrixQ[CZero],
            Function[boo] /.
                Table[
                    {boo -> SparseArray[ArrayRules[SparseArray[m]] /. nvar -> Slot[1], Dimensions[m]]},
                    {m, {CMinus, CZero, CPlus}}
                ],
            {CMinus, CZero, CPlus}
        ]
    ];

MatrixRSolve[{CMinus_, CZero_, CPlus_, cVector_, {L_, nvars_}}, {nvar_, n1_, n2_} /; n1 >= n2] :=
    Module[{matfuncs},
        matfuncs = MakeMatrixFunctions[{CMinus, CZero, CPlus}, nvar];
        aList[matfuncs, N[cVector], SList[matfuncs, n1], L, nvars, n2]
    ];

RecursionSolutionToReplacements[sol_, vars_, t_] :=
    Flatten[MapIndexed[MapThread[Rule, {vars /. t -> #2[[1]] - 1, #1}]&, sol]];

MatrixRSolve[recursmats_, {nvar_, n1_, n2_}, {vars_, t_}] :=
    RecursionSolutionToReplacements[MatrixRSolve[recursmats, {nvar, n1, n2}], vars, t];

(* ::Subsection::Closed:: *)
(*EquationsToMatrixRecursion old*)

(* ::Text:: *)
(*Convert a system of real, linear differential equations with explicit sinusoidal time dependence to the recursion form above.*)

(* ::Input:: *)
(*FourierExpandVariables[expr_,vars_,\[Omega]_,t_,n_]:=*)
(*Expand[TrigToExp[expr]]/.Dispatch@Flatten[{#->FE[#/.t->n,\[Omega],n],D[#,{t,j_}]->(I \[Omega] n)^j FE[#/.t->n,\[Omega],n]}&/@vars]*)

(* ::Input:: *)
(*IntegrateFourierTerms[polys_,vars_,\[Omega]_,t_,n_]:=*)
(*Replace[polys,x_?(FreeQ[#,FE]&)->x KroneckerDelta[n,0],{2}]/.{E^(Complex[0,j_] t \[Omega])  FE[a_,\[Omega],n]:>(a/.n->n-j), FE[a_,\[Omega],n]:>a};*)

(* ::Input:: *)
(*FourierExpandAndIntegrate[polys_,vars_,\[Omega]_,t_,n_]:=IntegrateFourierTerms[FourierExpandVariables[polys,vars,\[Omega],t,n],vars,\[Omega],t,n]/.n->0/.(((#/.t->0)->Re[#/.t->0])&/@vars)/.(((#/.t->(n_/;n<0))->Conjugate[#/.t->-n])&/@vars);*)

(* ::Input:: *)
(*EquationsToMatrixRecursion[eqs_,vars_,\[Omega]_,t_,n_]:=Module[{mat,varl},*)
(*varl=If[Head[vars]=!=List,{vars},vars];*)
(*mat=FourierExpandVariables[eqs/.Equal->Subtract,varl,\[Omega],t,n];*)
(*mat=IntegrateFourierTerms[mat,varl,\[Omega],t,n];*)
(*mat=Normal[CoefficientArrays[mat,Flatten[varl/.{{t->n},{t->n+1},{t->n-1}}]]];*)
(*Append[Transpose/@Partition[Transpose@mat[[2]],Length[varl]],mat[[1]]/.n->0]*)
(*]*)

(* ::Subsection:: *)
(*EquationsToMatrixRecursion with more frequencies*)

(* Write a variable or list of variables explicitly in terms of a specified number nmax of Fourier coefficients  *)

FourierExpandVariables[vars_List, w_, t_, nmax_Integer] :=
    ComplexExpand[
        Sum[(vars /. t -> n) E^(I n w t), {n, -nmax, nmax}] /.
            (((# /. t -> n_) /; n < 0 :> Conjugate[# /. t -> -n]) & /@ vars),
        vars /. t -> _ ,
        TargetFunctions -> {Re, Im}
    ] /. ((Im[# /. t -> 0] -> 0) & /@ vars);

FourierExpandVariables[vars_, w_, t_, nmax_Integer] :=
    First@FourierExpandVariables[{vars}, w, t, nmax];

FourierExpandVariables[expr_, vars_List, w_, t_, nmax_Integer] :=
    expr /. MapThread[Rule, {vars, FourierExpandVariables[vars, w, t, nmax]}];

FourierExpandVariables[expr_, vars_, w_, t_, nmax_Integer] :=
    FourierExpandVariables[expr, {vars}, w, t, nmax];

(*Convert a system of real, linear differential equations with explicit sinusoidal time dependence to the recursion form above.*)

FourierExpandVariables[expr_, vars_, \[Omega]_, t_, n_] :=
    TrigToExp[expr] //.
        (a_ + b_)c_ -> a c + b c /.
        Dispatch@Flatten[
            {# -> FE[# /. t -> n, \[Omega], n], D[#, {t, j_}] -> (I \[Omega] n)^j FE[# /. t -> n, \[Omega], n]}& /@ vars
        ];

IntegrateFourierTerms[polys_List, vars_, \[Omega]_, t_, n_] :=
    IntegrateFourierTerms[#, vars, \[Omega], t, n]& /@ polys;

IntegrateFourierTerms[polys_, vars_, \[Omega]_, t_, n_] :=
    Replace[polys, x_?(FreeQ[#, FE]&) -> x KroneckerDelta[n, 0], {1}] /.
        {E^(Complex[0, j_] t \[Omega] + c_.)  FE[a_, \[Omega], n] /; FreeQ[c, t] :> E^c (a /. n -> n - j), FE[a_, \[Omega], n] :> a};

FourierExpandAndIntegrate[polys_, vars_, \[Omega]_, t_] :=
    Module[{n},
        IntegrateFourierTerms[FourierExpandVariables[polys, vars, \[Omega], t, n], vars, \[Omega], t, n] /.
            n -> 0 /.
                (((# /. t -> 0) -> Re[# /. t -> 0])& /@ vars) /.
                    (((# /. t -> (m_ /; m < 0)) -> Conjugate[# /. t -> -m])& /@ vars)
    ];

Cmat[mats_, L_, n_, k_, m_] := If[Abs[k] > L, 0mats[[1]], mats[[k + L + 1]] /. n -> m];

CHat[mats_, L_, n_, k_, m_] := Module[{L2 = (L - 1) / 2},
    ArrayFlatten@Table[
        Cmat[mats, L, n, L k - q + r, L m + q],
        {q, -L2, L2},
        {r, -L2, L2}
    ]
];

cHat[cvec_, L_, n_] := Flatten@Table[KroneckerDelta[k]cvec /. n -> 0, {k, -(L - 1) / 2, (L - 1) / 2}];

EquationsToMatrixRecursion[eqs_, vars_, \[Omega]_, t_, n_] :=
    Module[{mat, varl, L, Cmats},
        varl = EnsureList@vars;
        mat = FourierExpandVariables[-eqs /. Equal -> Subtract, varl, \[Omega], t, n];
        mat = IntegrateFourierTerms[mat, varl, \[Omega], t, n];
        L = Max@Abs[Cases[mat, Alternatives @@ (varl /. t -> m_) -> m, \[Infinity]] - n];
        mat = Normal[CoefficientArrays[mat, Flatten[Table[varl /. t -> n + k, {k, -L, L}]]]];
        Cmats = Transpose /@ Partition[Transpose@mat[[2]], Length[varl]];
        Join[Table[CHat[Cmats, L, n, k, n + k], {k, -1, 1}], {cHat[mat[[1]], L, n]}, {{L, Length[vars]}}]
    ];

End[];

EndPackage[];
