(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`PlotAndFit`", {"GUIKit`", "AtomicDensityMatrix`"}];

If[$VersionNumber < 7., Needs["NonlinearRegression`"]];

(* ::Section:: *)
(*Implementation*)

(* ::Subsection:: *)
(*Begin private context*)

Begin["`Private`"];

(* ::Subsection:: *)
(*old PlotAndFit*)

Options[OldPlotAndFit] = Join[
    {DoFit -> True, ShowPlot -> True, FitParameters -> All, ReportErrors -> True},
    Options[Plot],
    Options[NonlinearRegress]
];

OldPlotAndFit[FitData_, FitFunction_, InitialParameters_, opts : OptionsPattern[]] :=
    Module[{FP, DummyFitVariables, FittedParameters},
        If[OptionValue[DoFit],
            FP = OptionValue[FitParameters];
            If[FP === All, FP = Table[True, {Length[InitialParameters]}]];
            DummyFitVariables = Table[If[FP[[i]], DummyVariable[i], InitialParameters[[i]]], {i, Length[InitialParameters]}];
            fit = NonlinearRegress[FitData, FitFunction[DummyFitVariables, x], Flatten[Table[If[FP[[i]], {{DummyVariable[i], InitialParameters[[i]]}}, {}], {i, Length[InitialParameters]}], 1], {x}, RegressionReport -> If[OptionValue[ReportErrors], {BestFitParameters, ParameterCITable}, {BestFitParameters}], Sequence @@ FilterRules[{opts}, Options@NonlinearRegress]];
            FittedParameters = DummyFitVariables /. (RegressionCommon`BestFitParameters /. fit);
            errors = If[OptionValue[ReportErrors], (RegressionCommon`ParameterCITable /. fit)[[1, All, 2]], 0InitialParameters];
            j = 1;errors1 = Table[If[FP[[i]], errors[[j++]], 0], {i, Length[FP]}],
            FittedParameters = InitialParameters;
            errors1 = 0InitialParameters
        ];
        If[OptionValue[ShowPlot],
            Print@Show[
                Plot[FitFunction[FittedParameters, x], {x, First[FitData][[1]], Last[FitData][[1]]}, PlotStyle -> RGBColor[1, 0, 0], DisplayFunction -> Identity],
                ListPlot[FitData, DisplayFunction -> Identity],
                DisplayFunction -> $DisplayFunction
            ]
        ];
        If[OptionValue[ReportErrors], Transpose[{FittedParameters, errors1}], FittedParameters]
    ];

(* ::Subsection:: *)
(*New PlotAndFit*)

Options[PlotAndFit] = Join[
    {DoFit -> True, ShowPlot -> True, HoldFixed -> {}, OutputHead -> List,
    Joined -> False, PlotStyle -> {Hue[0.67`, 0.6`, 0.6`], Red}},
    Options[Graphics],
    If[$VersionNumber < 7., Options[NonlinearRegress], Options[NonlinearModelFit]]
];

MakeVector[expr_] := If[VectorQ[expr], expr, {expr}];

MakeMatrix[expr_] := If[MatrixQ[expr], expr, {expr}];

EnsureList[expr_] := If[Head@expr === List, expr, {expr}];

PlotAndFit[data_, function_, params_, var_, opts : OptionsPattern[]] := Module[
    {fit, fixpos, params1, params2, fixrules},
    params1 = MakeMatrix[params];
    fixpos = Position[params1, Alternatives @@ ({#, _}& /@ MakeVector@OptionValue[HoldFixed]), {1}];
    params2 = Delete[params1, fixpos];
    fixrules = Apply[Rule, Extract[params1, fixpos], {1}];
    fit = If[ OptionValue[DoFit],
        If[$VersionNumber < 7.,
            BestFitParameters /. NonlinearRegress[
                data, function /. fixrules, params2, var,
                RegressionReport -> {BestFitParameters},
                Sequence @@ FilterRules[Join[{opts}, Options[PlotAndFit]], Options[NonlinearRegress]]
            ],
            NonlinearModelFit[
                data, function /. fixrules, params2, var,
                Sequence @@ FilterRules[Join[{opts}, Options[PlotAndFit]], Options[NonlinearModelFit]]
            ]["BestFitParameters"]
        ],
        Apply[Rule, params2, {1}]
    ];
    If[ OptionValue[ShowPlot],
        Print@Show[
            ListPlot[data,
                PlotStyle -> First@EnsureList@OptionValue[PlotStyle],
                Evaluate@FilterRules[{opts}, Options[ListPlot]]
            ],
            Plot[function /. fixrules /. fit, {var, Min[data[[All, 1]]], Max[data[[All, 1]]]},
                PlotStyle -> Last@EnsureList@OptionValue[PlotStyle],
                Evaluate@FilterRules[{opts}, Options@Plot]
            ],
            Evaluate@FilterRules[{opts}, {PlotRange}]
        ]
    ];
    Apply[OptionValue[OutputHead], Transpose[{params1[[All, 1]], params1[[All, 1]] /. fixrules /. fit}], {1}]
];

(* ::Subsection::Closed:: *)
(*Old GUIPlotAndFit*)

AllNumericQ[list_] := And @@ NumericQ /@ list;

NToString[x_] :=
    ToString[NumberForm[N[x], 4, ExponentFunction -> (If[-2 < # < 2, Null, #]&), NumberFormat -> (If[#3 == "", #1, SequenceForm[#1, "e", #3]]&)]];

yList[data_, func_, pnames_, args_?AllNumericQ] := Module[
    {func1},
    func1 = func[MapThread[Rule, {pnames, args}]];
    func1 /@ data[[All, 1]]
];

ChiSquare[data_, func_, pnames_, args_?AllNumericQ] := Total[(yList[data, func, pnames, args] - data[[All, 2]])^2];

DyList[i_, data_, func_, pnames_, args_] := (yList[data, func, pnames, MapAt[1.0001 #&, args, i]] - yList[data, func, pnames, args]) / (0.0001args[[i]]);

CovarianceMatrix[data_, func_, pnames_, args_, fitted_] := Module[{i, DyLists},
    DyLists = Flatten[Table[If[fitted[[i]], {DyList[i, data, func, pnames, args]}, {}], {i, Length[args]}], 1];
    ChiSquare[data, func, pnames, args]Inverse[Table[DyLists[[i]].DyLists[[j]], {i, Length[DyLists]}, {j, Length[DyLists]}]]
];

OldGUIPlotAndFit[data_, fitfunc_, params_] := Module[
    {numvals = Length[params], paramnames = params[[All, 1]], startparamlist, fitparamlist, errorlist, dofitlist, startvalorname, fitvalsandstart, updatePlot, makeplot, plotStarts, fit, didfit = False},
    GUIRunModal[Widget["Frame", {
        WidgetGroup[{
            Join[{{WidgetAlign[], Widget["Label", {"text" -> "Fit?"}],
                WidgetAlign[],
                Widget["Label", {"text" -> "Start value", "preferredSize" -> Widget["Dimension", {"width" -> 70, "height" -> 20}]}],
                WidgetAlign[],
                Widget["Label", {"text" -> "Fit value", "preferredSize" -> Widget["Dimension", {"width" -> 70, "height" -> 20}]}], WidgetAlign[], Widget["Label", {"text" -> "Error", "preferredSize" -> Widget["Dimension", {"width" -> 70, "height" -> 20}]}]}},
                MapIndexed[{Widget["Label", {"text" -> ToString[#[[1]]]}],
                    WidgetAlign[],
                    Widget["CheckBox", {"selected" -> If[Length[#] == 3, #[[3]] != 0, False]}, Name -> "doFit" <> ToString@#2[[1]]],
                    WidgetAlign[],
                    Widget["TextField", {"text" -> NToString[#[[2]]], BindEvent["action", Script[plotStarts[]]]}, Name -> "startVal" <> ToString@#2[[1]]], WidgetAlign[],
                    Widget["TextField", {"text" -> "", "editable" -> False}, Name -> "fitVal" <> ToString@#2[[1]]],
                    WidgetAlign[],
                    Widget["TextField", {"text" -> "", "editable" -> False}, Name -> "errorVal" <> ToString@#2[[1]]]
                }&, params]],
            {Widget["ImageLabel", {"preferredSize" -> Widget["Dimension", {"width" -> 350, "height" -> 300}]}, Name -> "canvas"]}
        },
            WidgetLayout -> {"Grouping" -> Row}
        ],
        {Widget["Button", {"text" -> "plot", BindEvent["action", Script[plotStarts[]]]}], Widget["Button", {"text" -> "fit", BindEvent["action", Script[If[Or @@ dofitlist[], fit[]]], InvokeThread -> "New"]}, Name -> "fitButton"],
        (*Widget["Button",{"text"->"abort","enabled"->False,BindEvent["action",Script[SetPropertyValue[{"abortButton","text"},"aborted",InvokeThread->"Dispatch"]],InvokeThread->"New"]},Name->"abortButton"],*)
            Widget["Button", {"text" -> "fit to start", BindEvent["action", Script[fittostart[]]]}],
            Widget["Button", {"text" -> "done", BindEvent["action", Script[InvokeMethod[{"frame", "dispose"}];]]}]
        },
        Script[
            startparamlist[] := Table[ImportString[PropertyValue[{"startVal" <> ToString[i], "Text"}, InvokeThread -> "Dispatch", InvokeWait -> True], "List"][[1]], {i, numvals}];
            fitparamlist[] := Table[ImportString[PropertyValue[{"fitVal" <> ToString[i], "Text"}, InvokeThread -> "Dispatch", InvokeWait -> True], "List"][[1]], {i, numvals}];
            errorlist[] := Table[ImportString[PropertyValue[{"errorVal" <> ToString[i], "Text"}, InvokeThread -> "Dispatch", InvokeWait -> True], "List"][[1]], {i, numvals}];
            dofitlist[] := Table[PropertyValue[{"doFit" <> ToString[i], "Selected"}, InvokeThread -> "Dispatch", InvokeWait -> True], {i, numvals}];
            startvalorname[] := MapThread[If[#1, #2, #3]&, {dofitlist[], paramnames, startparamlist[]}];
            fitvalsandstart[] := Flatten[MapThread[If[#1, {{#2, #3, #4}}, {}]&, {dofitlist[], paramnames, startparamlist[], 1.1startparamlist[]}], 1];
            updatePlot[plot_] := SetPropertyValue[{"canvas", "data"}, ExportString[plot, "GIF", ConversionOptions -> {"Transparency" -> GrayLevel[1]}, ImageSize -> 300], InvokeThread -> "Dispatch"];
            makeplot[p_] := Module[{func, plot, dataplot},
                func = fitfunc[p];
                Print[func];
                plot = Plot[func[x], {x, Min[data[[All, 1]]], Max[data[[All, 1]]]}, DisplayFunction -> Identity];
                dataplot = ListPlot[data, DisplayFunction -> Identity];
                Show[plot, dataplot, DisplayFunction -> Identity]
            ];
            plotStarts[] := updatePlot[makeplot[MapThread[Rule, {paramnames, startparamlist[]}]]];
            fit[] := Module[{fitresult, j, svon, dfl, fitparams, errors},
                j = 1;
                didfit = False;
                svon = startvalorname[];
                dfl = dofitlist[];
                SetPropertyValue[{"fitButton", "enabled"}, False, InvokeThread -> "Dispatch"];
                MapIndexed[SetPropertyValue[{"fitVal" <> ToString[#2[[1]]], "Text"}, "", InvokeThread -> "Dispatch"]&, svon];
                MapIndexed[SetPropertyValue[{"errorVal" <> ToString[#2[[1]]], "Text"}, "", InvokeThread -> "Dispatch"]&, svon];
                (*SetPropertyValue[{"abortButton","enabled"},True,InvokeThread->"Dispatch"];*)
                CheckAbort[
                    fitresult = FindMinimum[ChiSquare[data, fitfunc, paramnames, svon], fitvalsandstart[],
                        StepMonitor :> (
                            SetPropertyValue[{"fitButton", "text"}, j++, InvokeThread -> "Dispatch"];
                        (*MapIndexed[SetPropertyValue[{"fitVal"<>ToString[#2[[1]]],"Text"},NToString[#1],InvokeThread->"Dispatch"]&,svon];
updatePlot@makeplot@svon;*)
                        )];
                    fitparams = svon /. fitresult[[2]];
                    errors = Tr[CovarianceMatrix[data, fitfunc, paramnames, fitparams, dfl], List];
                    errors = Replace[svon, _?NumericQ -> 0, 1] /. Thread[Rule[fitvalsandstart[][[All, 1]], errors]];
                    MapIndexed[SetPropertyValue[{"fitVal" <> ToString[#2[[1]]], "Text"}, NToString[#1], InvokeThread -> "Dispatch"]&, fitparams];
                    MapIndexed[SetPropertyValue[{"errorVal" <> ToString[#2[[1]]], "Text"}, NToString[#1], InvokeThread -> "Dispatch"]&, errors];
                    updatePlot[makeplot[MapThread[Rule, {paramnames, fitparams}]]];
                    didfit = True,
                    $Aborted];
                SetPropertyValue[{"fitButton", "text"}, "fit", InvokeThread -> "Dispatch"];
                (*SetPropertyValue[{"abortButton","enabled"},False,InvokeThread->"Dispatch"];*)
                SetPropertyValue[{"fitButton", "enabled"}, True, InvokeThread -> "Dispatch"];
            ];
            fittostart[] := (MapIndexed[SetPropertyValue[{"startVal" <> ToString[#2[[1]]], "Text"}, NToString[#1], InvokeThread -> "Dispatch"]&, fitparamlist[]];);
            didfit = False;
            plotStarts[];
        ],
        BindEvent["endModal", Script[If[didfit, Show[makeplot[MapThread[Rule, {paramnames, fitparamlist[]}]], DisplayFunction -> $DisplayFunction];Transpose[{paramnames, fitparamlist[], errorlist[]}]]]]
    },
        Name -> "frame"
    ],
        IncludedScriptContexts -> {"Global`"}]
];

(* ::Subsection:: *)
(*New GUIPlotAndFit*)

FitAndErrorNR[data_, func_, params_, x_] := Module[{names = params[[All, 1]], fitvals, errors, fit},
    {fitvals, errors} = If[$VersionNumber < 7.,
        fit = NonlinearRegress[data, func, params, x];
        {names /. (BestFitParameters /. fit), (ParameterCITable /. fit)[[1, All, 2]]},
        fit = NonlinearModelFit[data, func, params, x];
        {names /. fit["BestFitParameters"], fit["ParameterErrors"]}
    ];
    MapThread[Rule, {names, Transpose[{fitvals, errors}]}]
];

GUIPlotAndFit[data_, func_, params_, x_] := DynamicModule[
    {params1, fitq, names, ivalues, fvalues, evalues, nv = Length[params],
        doFit = False, fitDone = False, setvals, newparams, setvals1},
    params1 = ReplacePart[Table[{0, 1, 1}, {nv}], Flatten[MapIndexed[#2 -> #1&, EnsureList /@ params, {2}]]];
    names = params1[[All, 1]];
    ivalues = params1[[All, 2]];
    fitq = (# =!= 0)& /@ params1[[All, 3]];
    fvalues = ivalues;
    evalues = 0fvalues;
    DialogInput[
        DialogNotebook[
            Panel[
                Grid[{{
                    Column[{
                        Grid[
                            Prepend[
                                Table[With[{i = i}, {
                                    names[[i]],
                                    Checkbox[Dynamic[fitq[[i]]]],
                                    InputField[Dynamic[ivalues[[i]]], FieldSize -> Small], InputField[Dynamic[fvalues[[i]]], FieldSize -> Small, Enabled -> False],
                                    InputField[Dynamic[evalues[[i]]], FieldSize -> Small, Enabled -> False]
                                }], {i, nv}],
                                {"", "Fit?", "Initial value", "Fit value", "Error"}
                            ]],
                        Row[{
                            Button["Plot", fitDone = False],
                            Button["Do Fit", doFit = True],
                            DefaultButton["Done", DialogReturn[Transpose[{names, fvalues}]]]
                        }]
                    }],
                    Dynamic[
                        If[doFit,
                            setvals = Delete[MapThread[Rule, {names, ivalues}], Position[fitq, True]];
                            newparams = Delete[Transpose[{names, ivalues}], Position[fitq, False]];
                            setvals1 = MapThread[Rule, {names, Transpose[{ivalues, Table[0, {nv}]}]}];
                            {fvalues, evalues} = Transpose[names /. FitAndErrorNR[data, func /. setvals, newparams, x] /. setvals1];
                            doFit = False;
                            fitDone = True
                        ];
                        Show[ListPlot[data], Plot[func /. MapThread[Rule, {names, If[fitDone, fvalues, ivalues]}], {x, Min[data[[All, 1]]], Max[data[[All, 1]]]}], ImageSize -> 72 5, Frame -> True, PlotRange -> All]
                    ]
                }}, Alignment -> Top]
            ]
        ]
    ]
];

End[];

EndPackage[];
