(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`Relaxation`", {"ProcessOptions`", "AtomicDensityMatrix`"}];

Begin["`Private`"];

IntrinsicRelaxation[sys_?AtomicSystemQ] = WignerEckart[sys, {NaturalWidth, 0, 0}];

TransitRelaxation[sys_?AtomicSystemQ, gamma_] := gamma WignerEckart[sys, {Identity, 0, 0}];

TransitRepopulation[sys_?AtomicSystemQ, gamma_] := Module[
    {g = SublevelMultiplicity@GroundStates@sys},
    gamma WignerEckart[sys, {GroundState, 0, 0}] / If[g == 0, 1, g]
];

Options[OpticalRepopulation] = Options@DMVariables;

OptionValidator[ExpandDipoleRME] = True | False;

(OpticalRepopulation[sys_?AtomicSystemQ, opts : OptionsPattern[]] :=
    Module[{representation},
        Switch[representation,
            {"Zeeman"},
            OpticalRepopulationZ[sys, FilterRules@opts],
            {"PolarizationMoments", ___},
            OpticalRepopulationPM[sys, Rest@representation, FilterRules@opts]
        ]
    ]) // ProcessOptions;

Options[OpticalRepopulationZ] = Options@OpticalRepopulation;

degeneracyBoole[energies_] := SparseArray@DirectSelfProduct[
    Composition[Boole, PossibleZeroQ, Subtract],
    energies
];

(OpticalRepopulationZ[sys_?AtomicSystemQ, opts : OptionsPattern[]] :=
    Module[{d, branchingMat, degenerateStates, dm, rank},
        (*spherical components of the electric dipole operator*)
        d = SparseArray@WignerEckart[sys, {"NormalizedDipole", 1}];
        (* Matrix of branching rates *)
        branchingMat = SparseArray@Sqrt[(2 J@sys + 1) NaturalWidth@sys DirectSelfProduct[BranchingRatio[#2][#1] &, sys]];
        (* Only allow transfer of coherences from a pair of degenerate levels to another pair of degenerate levels *)
        degenerateStates = degeneracyBoole[Energy@sys + HyperfineShift@sys];
        dm = DensityMatrix[sys, FilterRules@opts] degenerateStates;
        rank = Rank@d;
        Normal[degenerateStates *
            Sum[
                (-1)^q (Transpose@branchingMat Component[d, q]).dm.(branchingMat Component[d, -q]),
                {q, -rank, rank}
            ]
        ]
    ]) // ProcessOptions;

Options[OpticalRepopulationPM] = Options@PMVariables;

(OpticalRepopulationPM[sys_?AtomicSystemQ, opts : OptionsPattern[]] /; AtomicSystemType[sys] == "Toy" :=
    Map[{{{#}}}&, OpticalRepopulationZ[sys, Representation -> "Zeeman", FilterRules@opts], {2}]
) // ProcessOptions;

(OpticalRepopulationPM[sys_?AtomicSystemQ, opts : OptionsPattern[]] :=
    Module[{levels, excitedLevels, pme, ne, n, fe, f, repop, re, r, rate, hf},
        levels = RemoveZeemanSublevels@sys;
        excitedLevels = ExcitedStates@levels;
        pme = PolarizationMoments[excitedLevels, FilterRules@opts];
        repop = ZeroTensorDecompositionMatrix[sys, FilterRules@opts];
        {n, ne} = Length /@ {levels, excitedLevels};
        {f, fe} = F /@ {levels, excitedLevels};
        {r, re} = Diagonal /@ Rank /@ {repop, pme};
        rate = Table[(2 J@le + 1) NaturalWidth@le BranchingRatio[l][le], {l, levels}, {le, excitedLevels}];
        hf = If[AtomicSystemType@sys === "HyperfineZeeman",
            Table[hyperfineRMEFactor[l, 1, le] hyperfineRMEFactor[le, 1, l], {l, levels}, {le, excitedLevels}],
            ConstantArray[1, {Length@levels, Length@excitedLevels}]
        ];
        Do[
            repop[[i, i, k - r[[i, 1]] + 1]] +=
                (-1)^(2 f[[i]] + k) hf[[i, j]] rate[[i, j]] Component[pme[[j, j]], k] SJS[{k, f[[i]], f[[i]]}, {1, fe[[j]], fe[[j]]}],
            {i, n},
            {j, ne},
            {k, Max[r[[i, 1]], re[[j, 1]]], Min[r[[i, -1]], re[[j, -1]]]}
        ];
        repop
    ]) // ProcessOptions;

OpticalRepopulationOperator[sys_?AtomicSystemQ] :=
    Module[{dm, or},
        dm = DensityMatrix[sys, ComplexExpandVariables -> False, Representation -> "Zeeman"];
        or = OpticalRepopulation[sys, ComplexExpandVariables -> False];
        Normal@Transpose@Map[Partition[#, Length[sys]]&, Last@CoefficientArrays[or, Flatten@dm], {2}]
    ];

(*The OpticalRepopulation matrix is equal to Map[Tr[#\[CenterDot]DensityMatrix[System]]&,OpticalRepopulationOperator[System],{2}]*)

Relaxation[as_?AtomicSystemQ, es : {_ExperimentalRegion..}] := (
    IntrinsicRelaxation[as] +
        TransitRelaxation[as, TransitRate[VoidRegion][#]] +
            TransitRelaxation[as, Sum[TransitRate[l][#], {l, es}]]
)& /@ es;

Options@Repopulation = Join[
    {Method -> Automatic},
    FilterRules[Options@DMVariables, Except@DMLabel]
];

(Repopulation[as_?AtomicSystemQ, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    (TransitRepopulation[as, Density@# TransitRate[VoidRegion][#]] +
        OpticalRepopulation[as, DMLabel -> RegionLabel@#, FilterRules@opts] +
        Sum[
            Density@# / Density@l TransitRate[l][#] *
                DensityMatrix[as, DMLabel -> RegionLabel@l, FilterRules@opts],
            {l, es}
        ]
    )& /@ es
) // ProcessOptions;

FluorescenceOperator[sys_?AtomicSystemQ, freq_, th_, phi_] :=
    Module[{dc, gr, ikk},
        If[AtomicSystemType[sys] == "Toy", Return[FluorescenceOperator[sys, freq] / (4 Pi)]];
        dc = ExpandDipoleRME[sys, ToCartesian@WignerEckart[sys, {Dipole, 1}]];
        gr = DiagonalMatrix[Boole /@ GroundStateQ /@ sys];
        ikk = IdentityMatrix@3 - DirectSelfProduct[{Cos[phi] Sin[th], Sin[phi] Sin[th], Cos[th]}];
        freq^4 / (2\[Pi]) Sum[dc[[i]].gr.dc[[j]] ikk[[i, j]], {i, 3}, {j, 3}]
    ];

FluorescenceOperator[sys_?AtomicSystemQ, freq_] :=
    Module[{dc, gr},
        dc = ExpandDipoleRME[sys, ToCartesian@WignerEckart[sys, {Dipole, 1}]];
        gr = DiagonalMatrix[Boole /@ GroundStateQ /@ sys];
        4 / 3 freq^4 Sum[dc[[i]].gr.dc[[i]], {i, Length@dc}]
    ];

Options[Fluorescence] = Options@DMVariables;

branchingRatioMatrix[sys_] := SparseArray@Sqrt[
    (2 J[sys] + 1) NaturalWidth[sys] DirectSelfProduct[BranchingRatio[#2][#1] &, sys]
];

energyDifferenceMatrix[sys_] := SparseArray@Sqrt@DirectSelfProduct[Subtract, Energy@sys];

degeneracyMatrix[sys_] :=
    SparseArray@DirectSelfProduct[
        Composition[Boole, PossibleZeroQ, Subtract],
        Energy@sys + HyperfineShift@sys
    ];

(*Fluorescence[sys_?AtomicSystemQ, freq_, th_, phi_, opts___Rule] :=*)
(*Tr[DensityMatrix[sys].FluorescenceOperator[sys, freq, th, phi, opts]];*)

Fluorescence[sys_?AtomicSystemQ, th_, phi_, opts : OptionsPattern[]] /; AtomicSystemType[sys] == "Toy" :=
    1 / (4 Pi) Fluorescence[sys, opts];

(Fluorescence[sys_?AtomicSystemQ, th_, phi_, opts : OptionsPattern[]] :=
    Module[{d, bMat, dm, ikk},
        d = SparseArray@ToCartesian@WignerEckart[sys, {"NormalizedDipole", 1}];
        bMat = energyDifferenceMatrix@sys branchingRatioMatrix@sys;
        dm = DensityMatrix[sys, FilterRules@opts] degeneracyMatrix@sys;
        ikk = IdentityMatrix[3] - DirectSelfProduct[{Cos[phi] Sin[th], Sin[phi] Sin[th], Cos[th]}];
        3 / (8 Pi) Sum[Tr[(Transpose@bMat Component[d, i]).dm.(bMat Component[d, j])] ikk[[i, j]], {i, 3}, {j, 3}]
    ]) // ProcessOptions;

(Fluorescence[sys_?AtomicSystemQ, opts : OptionsPattern[]] :=
    Module[{d, bMat, dm, rank},
        d = SparseArray@WignerEckart[sys, {"NormalizedDipole", 1}];
        rank = Rank@d;
        bMat = energyDifferenceMatrix@sys branchingRatioMatrix@sys;
        dm = DensityMatrix[sys, FilterRules@opts] degeneracyMatrix@sys;
        Sum[(-1)^q Tr[(Transpose@bMat Component[d, q]).dm.(bMat Component[d, -q])], {q, -rank, rank}]
    ]) // ProcessOptions;

End[];

EndPackage[];
